//---------------------------------------------------
// module  : f_GCEBFM.cc                       
// auteur  : Lazhar Khelifi and Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "GCEBFM.h"

//------------------------------------------------
// METHODES DE LA CLASSE GCEBFM  ------------------
//------------------------------------------------
//----------------------------------------------------------
// Constructeur de la classe GCEBFM
//----------------------------------------------------------
GCEBFM::GCEBFM(int*** imgt,int lgth,int wdth)           
 {
  int i,j,k;

  //Initialisation
  length=lgth;
  width=wdth;

  //Allocation Memoire Matrice 
  mat_img=imatrix_allocate_3d(3,length,width);
  mat_cnt=imatrix_allocate_3d(3,length,width);
  mat_seg=fmatrix_allocate_2d(length,width);

  //Initialisation
  for(k=0;k<3;k++)
  for(i=0;i<length;i++) for(j=0;j<width;j++) 
    { mat_img[k][i][j]=imgt[k][i][j];     
      mat_cnt[k][i][j]=imgt[k][i][j]; }

 //Stack Size Increased (128 MB)
 const rlim_t kStackSize=128*1024*1024;   
 struct rlimit rl;
 int result=getrlimit(RLIMIT_STACK,&rl);
 if (result==0)
 if (rl.rlim_cur<kStackSize)
    {  rl.rlim_cur=kStackSize;
       result=setrlimit(RLIMIT_STACK,&rl);
       if (result!=0) fprintf(stderr,"setrlimit returned result = %d\n",result); } 
 }

//----------------------------------------------------------
// Destructeur de la classe VOIFBM
//----------------------------------------------------------
GCEBFM::~GCEBFM()
{ 
  //Libere Memoire Matrice 
  if (mat_img)    free_imatrix_3d(mat_img,3);
  if (mat_cnt)    free_imatrix_3d(mat_cnt,3);
  if (mat_seg)    free_fmatrix_2d(mat_seg);
}

//----------------------------//
//----------GESTION-----------//
//----------------------------//
//----------------------------------------------------------
// MultipleSegmentations
// ---------------------
//
//----------------------------------------------------------  
void GCEBFM::MultipleSegmentations(int*** TabSeg,int NbSpCl,int NbCl1,int NbCl2,int NbCl3)
{
 int k,l,m;
 int nbreg;

 //Presentation
 printf("\n\n  > MultipleSegmentations [NbSpCol::%d][NbBins::%d][Size::%d]",NbSpCl,NBINS,SIZE1);
 fflush(stdout);

 //-------//
 // Coul  //
 //-------//
   printf("\n  [%d]<NbCl=%d><Coul>",NbSpCl,NbCl1);
   for(k=0,l=0;k<NbSpCl;k++,l++) 
     { SegmentationHistoCOLORKmeanFast(mat_img,TabSeg[k],length,width,NbCl1,NBINS,SIZE1,k,l);
       for(m=0;m<5;m++) if (!FuseSmallRegions(TabSeg[k],length,width,&nbreg,SIZEMIN))  break; } 

   printf("\n  [%d]<NbCl=%d><Coul>",NbSpCl,NbCl2);
   for(k=NbSpCl,l=0;k<(2*NbSpCl);k++,l++) 
     { SegmentationHistoCOLORKmeanFast(mat_img,TabSeg[k],length,width,NbCl2,NBINS,SIZE1,k,l);
       for(m=0;m<5;m++) if (!FuseSmallRegions(TabSeg[k],length,width,&nbreg,SIZEMIN))  break; } 

   printf("\n  [%d]<NbCl=%d><Coul>",NbSpCl,NbCl3);
   for(k=(2*NbSpCl),l=0;k<(3*NbSpCl);k++,l++) 
      { SegmentationHistoCOLORKmeanFast(mat_img,TabSeg[k],length,width,NbCl3,NBINS,SIZE1,k,l);
        for(m=0;m<5;m++) if (!FuseSmallRegions(TabSeg[k],length,width,&nbreg,SIZEMIN))  break; } 


 //-------//
 // NbBin //
 //-------//
   printf("\n  [%d]<NbCl=%d><Nbins-1>",NbSpCl,NbCl1);
   for(k=(3*NbSpCl),l=0;k<(4*NbSpCl);k++,l++) 
     { SegmentationHistoCOLORKmeanFast(mat_img,TabSeg[k],length,width,NbCl1,(NBINS-1),SIZE1,k,l);
       for(m=0;m<5;m++) if (!FuseSmallRegions(TabSeg[k],length,width,&nbreg,SIZEMIN))  break; }  

   printf("\n  [%d]<NbCl=%d><Nbins-1>",NbSpCl,NbCl2);
   for(k=(4*NbSpCl),l=0;k<(5*NbSpCl);k++,l++) 
     { SegmentationHistoCOLORKmeanFast(mat_img,TabSeg[k],length,width,NbCl2,(NBINS-1),SIZE1,k,l);
       for(m=0;m<5;m++) if (!FuseSmallRegions(TabSeg[k],length,width,&nbreg,SIZEMIN))  break; }  
}

//===============
//== GCE-BFM ====
//===============
//----------------------------------------------------------
// GCE-BasedFusionModel
//----------------------------------------------------------    
void GCEBFM::GCEFusionModel()
{
 int k;
 int nbreg;
 float dist;
 char MyBuff[100]; 
 int IndxBestSegGCE;

 //>Record
 strcpy(MyBuff,Extension);
 
 //>Information
 printf("\n\n GCE-BFM Algorithm");
 printf("\n ------------------");
 printf("\n Image: [%s] (%d - %d)",Name_Img,length,width);
 printf("\n ---------");
 //printf("\n  >> Regul=%.4f",Regularization);
 //printf("\n --------");
 printf("\n  >> Regul_Alpha=%.3f",RegulAlpha);
 printf("\n  >> Lg_Mean_Reg=%.3f",LogMeanReg);
 printf("\n -----------\n");
 fflush(stdout);

 //>Parametre 
 float P_REGULALPHA=RegulAlpha; 
 float P_LOG_MEAN_REG=LogMeanReg;

 //>Constantes
 int NBSEGS=5*NBSPCOL;
 
 //>MemoryAllocation
 int****  PopSegInit=imatrix_allocate_4d(2,NBSEGS,length,width);
 float*   TabDist=fmatrix_allocate_1d(NBSEGS);
 float**  Mat1=fmatrix_allocate_2d(length,width);
 int**    MatInt1=imatrix_allocate_2d(length,width);
 int**    MatInt2=imatrix_allocate_2d(length,width);
 int**    MatInt3=imatrix_allocate_2d(length,width);

 //>Complexity
 float complexity=Complexity(mat_img,length,width,5,7);

 
 //===============================================================================
 //> [0] ESTIMATION-<Nb.Classes>
 //===============================================================================
 int NbCl1=(int)(0.5+(NBCL1*complexity));  
 int NbCl2=(int)(0.5+(NBCL2*complexity));
 int NbCl3=(int)(0.5+(NBCL3*complexity));
 if (NbCl1<2) NbCl1=2;
 if (NbCl2<2) NbCl2=2;
 if (NbCl3<2) NbCl3=2;
 printf(" :: [%f] >> [NbCl1:%d][NbCl2:%d][NbCl3:%d]",complexity,(int)NbCl1,(int)NbCl2,(int)NbCl3);
 fflush(stdout);


 //================================================================================
 //> [1] Init. Population :: Multiple ColorSegmentationsRegions [0:class][1:regs]
 //================================================================================
 //>Multiple ColorSegRegions
 MultipleSegmentations(PopSegInit[0],NBSPCOL,NbCl1,NbCl2,NbCl3);
 for(int k=0;k<NBSEGS;k++) ConvertClassRegionInf(PopSegInit[0][k],PopSegInit[1][k],length,width);

 //DEBUG
 //SaveTabImagePgm("","ArrayKmeans",PopSegInit[1],length,width,NBSEGS);;

 //>Classification & Best->MatInt1
 IndxBestSegGCE=FM_SelectSegMapGCE(PopSegInit[1],length,width,NBSEGS,TabDist,&dist,P_REGULALPHA,P_LOG_MEAN_REG);
 Classify(PopSegInit[1],length,width,NBSEGS,TabDist,1); 
 //RandomImgHor(PopSegInit[1][0],length,width,5); 
 //CopyMat(MatInt1,PopSegInit[1][(int)Regularization],length,width);
 CopyMat(MatInt1,PopSegInit[1][0],length,width);  

 //DEBUG
 //SaveTabImagePgm("","ArrayKmeansClassified",PopSegInit[1],length,width,NBSEGS);

 //===============================================================================
 //> [2] ICM_GCEBFM
 //===============================================================================
 FM_ICM_GCESupPix(MatInt1,MatInt2,PopSegInit[1],length,width,NBSEGS,NBITERICM,P_REGULALPHA,PRECIS,P_LOG_MEAN_REG);

 //DEBUG
 //SaveImagePgm("","SegInit",MatInt1,length,width);
 //SaveImagePgm("","SegFin",MatInt2,length,width);

  
 //===============================================================================
 //> [3] Clean & Visu
 //===============================================================================
 //>Clean
 for(k=0;k<2;k++) if (!FuseSmallRegions(MatInt2,length,width,&nbreg,SIZEMIN/2))  break; 
 CopyMat(Mat1,MatInt2,length,width);
 for(k=0;k<2;k++) if (!FuseSmallRegions(Mat1,length,width,&nbreg,SIZEMIN/2,0.6))  break;
 CopyMat(MatInt2,Mat1,length,width);

 //>Classe->Regions
 ConvertClassRegionInf(MatInt2,MatInt3,length,width);
 //printf("\n   > GCE=%.3f",FM_ComputeGCE(MatInt3,PopSegInit[1],length,width,NBSEGS,0.0,0.0)); 
 fflush(stdout);
 Recal(MatInt3,length,width);
 
 //>Save
 if (SAVE) SaveImagePgm(Name_Img,(char*)"_Z00",MatInt3,length,width);

 //>Visu
 AddContourImg(Mat1,mat_cnt,length,width,3);
 CopyMat(mat_seg,MatInt3,length,width);
 SaveImagePpm(Name_Img,(char*)"_Z00_CONT",mat_cnt,length,width);
 

 //> END
 //===============================================================================
 //>FreeMemory 
 if (Mat1)     free_fmatrix_2d(Mat1);
 if (MatInt1)  free_imatrix_2d(MatInt1);
 if (MatInt2)  free_imatrix_2d(MatInt2);
 if (MatInt3)  free_imatrix_2d(MatInt3);
 if (TabDist)  free_fmatrix_1d(TabDist); 
}
